import mlpack
import scipy.io as sio
import numpy as np
import matplotlib.pyplot as plt
from flcode import *

# function to calculate MST cost using disance matrix
def mst_cost(distances, edges):
    m = edges.shape[0]
    cost = 0.0
    for i in range(m):
        start_vtx = edges[i, 0]
        end_vtx = edges[i, 1]
        cost += distances[int(start_vtx), int(end_vtx)]
    return cost


# Load data
print('loading data')
mat_contents = sio.loadmat('data/face_data.mat')
face_dataset = mat_contents['images'].T
distance_face = distance_matrix(face_dataset, face_dataset)     
distance_sorted_face = np.copy(distance_face)
distance_sorted_face.sort(axis=1)
print('data loaded')

# Running time experiments
print('doing time experiment')
print('NOTE: make sure you are not running any other intensive program since that can affect time calculations')

time_d = []
for d in range(5,21):
    curr_time = np.zeros(50)
    for j in range(50):
        start = time.time()
        reduced_points_face = do_random_projection(face_dataset, d)
        mlpack.emst(reduced_points_face)
        curr_time[j] = time.time()-start
    time_d.append(np.median(curr_time))

# Cost
print('doing cost experiment')
cost_d = []
for d in range(5,21):
    curr_cost = 0.0
    for j in range(20):
        reduced_points_face = do_random_projection(face_dataset, d)
        output = mlpack.emst(reduced_points_face)['output']
        curr_cost += mst_cost(distance_face, output[:, 0:2])*.05
    cost_d.append(curr_cost)

# Actual cost
output = mlpack.emst(face_dataset)['output']
act_cost = output[:, 2].sum()

# Figure 4a of the paper
print('plotting results')

plt.rcParams['ytick.major.pad'] = 0

fig, ax1 = plt.subplots(figsize=(10,8))
plt.rc('font', family='serif')

                        
ax1.set_xlabel("Projected Dimension", fontsize=20)
ax1.set_ylabel("Pullback Tree / MST", fontsize=20,color="C0", fontweight="bold")


ax1.plot(range(5,21), [k/act_cost for k in cost_d], '.-', label = "Ratio", markersize=12, linewidth=1.5,color="C0")

ax1.tick_params(labelsize=20)
ax1.tick_params(axis='y', labelcolor="C0")
ax1.set_title("Faces Dataset", fontsize=20)

ax2 = ax1.twinx()
ax2.set_ylabel("Average Running Time (s)", fontsize=20, color="C1", fontweight="bold")

ax2.plot(range(5,21), time_d, '.-', label = "MST Time", markersize=12, linewidth=1.5, color="C1")


ax2.tick_params(labelsize=20)
ax2.tick_params(axis='y', labelcolor="C1")

fig.tight_layout()   

for label in ax1.get_yticklabels():
    label.set_fontweight("bold")
    
for label in ax2.get_yticklabels():
    label.set_fontweight("bold")
# plt.savefig("mst_faces.pdf", bbox_inches = 'tight', pad_inches = 0)
plt.show()